from typing import Optional
from decimal import Decimal
from fractions import Fraction

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.animation import AnimationOptions


class PatternOptions(HighchartsMeta):
    """Definition of a pattern."""

    def __init__(self, **kwargs):
        self._aspect_ratio = None
        self._background_color = None
        self._color = None
        self._height = None
        self._id = None
        self._image = None
        self._opacity = None
        self._path = None
        self._pattern_transform = None
        self._width = None
        self._x = None
        self._y = None

        self.aspect_ratio = kwargs.get('aspect_ratio', None)
        self.background_color = kwargs.get('background_color', None)
        self.color = kwargs.get('color', None)
        self.height = kwargs.get('height', None)
        self.id = kwargs.get('id', None)
        self.image = kwargs.get('image', None)
        self.opacity = kwargs.get('opacity', None)
        self.path = kwargs.get('path', None)
        self.pattern_transform = kwargs.get('pattern_transform', None)
        self.width = kwargs.get('width', None)
        self.x = kwargs.get('x', None)
        self.y = kwargs.get('y', None)

    @property
    def aspect_ratio(self) -> Optional[int | float | Decimal | Fraction]:
        """For automatically calculated width and height on images, it is possible to set
        an aspect ratio. The image will be zoomed to fill the bounding box, maintaining
        the aspect ratio defined.

        :returns: The aspect ratio applied.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._aspect_ratio

    @aspect_ratio.setter
    def aspect_ratio(self, value):
        self._aspect_ratio = validators.numeric(value, allow_empty = True)

    @property
    def background_color(self) -> Optional[str]:
        """Background color for the pattern if a :meth:`PatternOptions.path` is set (not
        images).

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        self._background_color = validators.string(value, allow_empty = True)

    @property
    def color(self) -> Optional[str]:
        """Pattern color, used as default path stroke.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        self._color = validators.string(value, allow_empty = True)

    @property
    def height(self) -> Optional[int | float | Decimal]:
        """Height of the pattern expressed in pixels.

        .. note::

          For images this is automatically set to the height of the element bounding box
          if not supplied.

          For non-image patterns the default is ``32px``.

          Note that automatic resizing of image patterns to fill a bounding box
          dynamically is only supported for patterns with an automatically calculated ID.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._height

    @height.setter
    def height(self, value):
        self._height = validators.numeric(value, allow_empty = True)

    @property
    def id(self) -> Optional[str]:
        """ID to assign to the pattern.

        .. note::

          This is automatically computed if not specified explicitly, and identical
          patterns are reused.

          To refer to an existing pattern for a Highcharts color in JavaScript, use color:
          ``"url(#pattern-id)"``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._id

    @id.setter
    def id(self, value):
        self._id = validators.string(value, allow_empty = True)

    @property
    def image(self) -> Optional[str]:
        """URL to an image to use as the pattern.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises ValueError: if value provided is not URL or path-like
        """
        return self._image

    @image.setter
    def image(self, value):
        try:
            self._image = validators.url(value, allow_empty = True)
        except ValueError:
            self._image = validators.path(value, allow_empty = True)

    @property
    def opacity(self) -> Optional[float]:
        """Opacity of the pattern as a float value from 0 to 1.

        :rtype: :class:`float <python:float` or :obj:`None <python:None>`

        :raises ValueError: if outside the range 0 - 1
        """
        return self._opacity

    @opacity.setter
    def opacity(self, value):
        self._opacity = validators.float(value,
                                         allow_empty = True,
                                         minimum = 0,
                                         maximum = 1)

    @property
    def path(self) -> Optional[str]:
        """An SVG path expressed as a string.

        .. note::

          If a path is supplied for the pattern, the :meth:`PatternOptions.image` property
          is ignored.

        .. warning::

          Highcharts for Python does not yet support the JavaScript
          ``Highcharts.SVGAttributes`` object. Only strings are currently supported.

        .. todo:

          Add ``Highcharts.SVGAttributes`` to utility classes.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._path

    @path.setter
    def path(self, value):
        self._path = validators.string(value, allow_empty = True)

    @property
    def pattern_transform(self) -> Optional[str]:
        """SVG ``patternTransform`` to apply to the entire pattern.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._pattern_transform

    @pattern_transform.setter
    def pattern_transform(self, value):
        self._pattern_transform = validators.string(value, allow_empty = True)

    @property
    def width(self) -> Optional[int | float | Decimal]:
        """Width of the pattern expressed in pixels.

        .. note::

          For images this is automatically set to the width of the element bounding box
          if not supplied.

          For non-image patterns the default is ``32px``.

          Note that automatic resizing of image patterns to fill a bounding box
          dynamically is only supported for patterns with an automatically calculated ID.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = validators.numeric(value, allow_empty = True)

    @property
    def x(self) -> Optional[int | float | Decimal]:
        """Horizontal offset applied to the pattern. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._x

    @x.setter
    def x(self, value):
        self._x = validators.numeric(value, allow_empty = True)

    @property
    def y(self) -> Optional[int | float | Decimal]:
        """Vertical offset applied to the pattern. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._y

    @y.setter
    def y(self, value):
        self._y = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'aspect_ratio': as_dict.get('aspectRatio', None),
            'background_color': as_dict.get('backgroundColor', None),
            'color': as_dict.get('color', None),
            'height': as_dict.get('height', None),
            'id': as_dict.get('id', None),
            'image': as_dict.get('image', None),
            'opacity': as_dict.get('opacity', None),
            'path': as_dict.get('path', None),
            'pattern_transform': as_dict.get('patternTransform', None),
            'width': as_dict.get('width', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None)
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'aspectRatio': self.aspect_ratio,
            'backgroundColor': self.background_color,
            'color': self.color,
            'height': self.height,
            'id': self.id,
            'image': self.image,
            'opacity': self.opacity,
            'path': self.path,
            'patternTransform': self.pattern_transform,
            'width': self.width,
            'x': self.x,
            'y': self.y
        }

        return untrimmed


class Pattern(HighchartsMeta):
    """Holds a pattern definition."""

    def __init__(self, **kwargs):
        self._animation = None
        self._pattern_options = None
        self._pattern_index = None

        self.animation = kwargs.get('animation', None)
        self.pattern_options = kwargs.get('pattern_options', None)
        self.pattern_index = kwargs.get('pattern_index', None)

    @property
    def animation(self) -> Optional[AnimationOptions]:
        """Animations options for the loading of image patterns.

        :returns: Settings for the animation of image patterns.
        :rtype: :class:`AnimationOptions` or :obj:`None <python:None>`
        """
        return self._animation

    @animation.setter
    @class_sensitive(AnimationOptions)
    def animation(self, value):
        self._animation = value

    @property
    def pattern_options(self) -> Optional[PatternOptions]:
        """Options to define a Pattern.

        :rtype: :class:`PatternOptions` or :obj:`None <python:None>`
        """
        return self._pattern_options

    @pattern_options.setter
    @class_sensitive(PatternOptions)
    def pattern_options(self, value):
        self._pattern_options = value

    @property
    def pattern_index(self) -> Optional[int]:
        """An optional index that indicates which of Highcharts default patterns to apply.

        Default patterns are part of the (JavaScript) ``Highcharts.patterns`` array, and
        additional patterns may be added to that array as well.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._pattern_index

    @pattern_index.setter
    def pattern_index(self, value):
        self._pattern_index = validators.integer(value,
                                                 allow_empty = True,
                                                 minimum = True,
                                                 coerce_value = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        pattern_options = as_dict.get('pattern', None) or \
                          as_dict.get('pattern_options', None) or \
                          as_dict.get('patternOptions', None)
        kwargs = {
            'animation': as_dict.get('animation', None),
            'pattern_options': pattern_options,
            'pattern_index': as_dict.get('patternIndex', None)
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'animation': self.animation,
            'pattern': self.pattern_options,
            'patternIndex': self.pattern_index
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core.options.series.bar import BarSeries
from highcharts_core.options.plot_options.histogram import HistogramOptions
from highcharts_core.utility_functions import mro__to_untrimmed_dict


class HistogramSeries(BarSeries, HistogramOptions):
    """Options to configure a Histogram series.

    A histogram is a column series which represents the distribution of the data set
    in the base series. Histogram splits data into bins and shows their frequencies.

    .. figure:: ../../../_static/histogram-example.png
      :alt: Histogram Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._base_series = None

        self.base_series = kwargs.get('base_series', None)

        super().__init__(**kwargs)

    @property
    def base_series(self) -> Optional[int | str]:
        """An integer identifying the index to use for the base series, or a string
        representing the id of the series. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`int <python:int>` or :class:`str <python:str>` or
          :obj:`None <python:None>`
        """
        return self._base_series

    @base_series.setter
    def base_series(self, value):
        if value is None:
            self._base_series = None
        else:
            try:
                value = validators.string(value)
            except (TypeError, ValueError):
                value = validators.integer(value,
                                           minimum = 0)

            self._base_series = value

    @property
    def data(self) -> None:
        """The collection of data points for the series. Defaults to
        :obj:`None <python:None>`.

        .. warning::

          All Histogram Series by definition return :obj:`None <python:None>` for their
          data. They are a special series, drawn in relationship to the
          :meth:`base_series <HistogramSeries.base_series>` specified, and do not receive
          independent data points of their own.

        :rtype: :obj:`None <python:None>`
        """
        return None

    @data.setter
    def data(self, value):
        pass

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'fill_color': as_dict.get('fillColor', None),
            'fill_opacity': as_dict.get('fillOpacity', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_color': as_dict.get('lineColor', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'negative_fill_color': as_dict.get('negativeFillColor', None),
            'point_interval': as_dict.get('pointInterval', None),
            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'track_by_area': as_dict.get('trackByArea', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'center_in_category': as_dict.get('centerInCategory', None),
            'color_by_point': as_dict.get('colorByPoint', None),
            'colors': as_dict.get('colors', None),
            'grouping': as_dict.get('grouping', None),
            'group_padding': as_dict.get('groupPadding', None),
            'max_point_width': as_dict.get('maxPointWidth', None),
            'min_point_length': as_dict.get('minPointLength', None),
            'point_padding': as_dict.get('pointPadding', None),
            'point_range': as_dict.get('pointRange', None),
            'point_width': as_dict.get('pointWidth', None),

            'depth': as_dict.get('depth', None),
            'edge_color': as_dict.get('edgeColor', None),
            'edge_width': as_dict.get('edgeWidth', None),
            'group_z_padding': as_dict.get('groupZPadding', None),

            'data': as_dict.get('data', None),
            'id': as_dict.get('id', None),
            'index': as_dict.get('index', None),
            'legend_index': as_dict.get('legendIndex', None),
            'name': as_dict.get('name', None),
            'stack': as_dict.get('stack', None),
            'x_axis': as_dict.get('xAxis', None),
            'y_axis': as_dict.get('yAxis', None),
            'z_index': as_dict.get('zIndex', None),

            'bins_number': as_dict.get('binsNumber', None),
            'bin_width': as_dict.get('binWidth', None),

            'base_series': as_dict.get('baseSeries', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'baseSeries': self.base_series
        }

        parents_as_dict = mro__to_untrimmed_dict(self, in_cls = in_cls)
        for key in parents_as_dict:
            untrimmed[key] = parents_as_dict[key]

        return untrimmed



"""Implements the Python representation of the Accessibility module."""
from typing import Optional, Any
from copy import deepcopy

from validator_collection import validators

from highcharts_core.metaclasses import HighchartsMeta, JavaScriptDict
from highcharts_core.decorators import class_sensitive
from highcharts_core import constants, errors

from highcharts_core.options.accessibility.announce_new_data import AnnounceNewData
from highcharts_core.options.accessibility.high_contrast_theme import HighContrastTheme
from highcharts_core.options.accessibility.keyboard_navigation import KeyboardNavigation
from highcharts_core.options.accessibility.point import AccessibilityPoint
from highcharts_core.options.accessibility.screen_reader_section import ScreenReaderSection
from highcharts_core.options.accessibility.series import SeriesAccessibility


class CustomAccessibilityComponents(JavaScriptDict):
    """JavaScript object which contains a map of component names to instances of classes
    inheriting from the (JavaScript)
    `Highcharts.AccessibilityComponent <https://api.highcharts.com/class-reference/Highcharts.AccessibilityComponent>`_
    base class.

    .. tip::

      Remember to add the component to the :meth:`keyboard_navigation.order`
      for the keyboard navigation to be usable.

    """
    _valid_value_types = str
    _allow_empty_value = True


class Accessibility(HighchartsMeta):
    """Options for configuring accessibility for the chart."""

    def __init__(self, **kwargs):
        self._announce_new_data = None
        self._custom_components = None
        self._description = None
        self._enabled = True
        self._high_contrast_theme = None
        self._keyboard_navigation = None
        self._landmark_verbosity = None
        self._linked_description = None
        self._point = None
        self._screen_reader_section = None
        self._series = None
        self._type_description = None

        self.announce_new_data = kwargs.get('announce_new_data', None)
        self.custom_components = kwargs.get('custom_components', None)
        self.description = kwargs.get('description', None)
        self.enabled = kwargs.get('enabled', None)
        self.high_contrast_theme = kwargs.get('high_contrast_theme', None)
        self.keyboard_navigation = kwargs.get('keyboard_navigation', None)
        self.landmark_verbosity = kwargs.get('landmark_verbosity', None)
        self.linked_description = kwargs.get('linked_description', None)
        self.point = kwargs.get('point', None)
        self.screen_reader_section = kwargs.get('screen_reader_section', None)
        self.series = kwargs.get('series', None)
        self.type_description = kwargs.get('type_description', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'accessibility'

    @property
    def announce_new_data(self) -> Optional[AnnounceNewData]:
        """Options for announcing new data to screen reader users.

        .. tip::

          Useful for dynamic data applications and drilldown.

          Keep in mind that frequent announcements will not be useful to users, as they
          won't have time to explore the new data. For these applications, consider making
          snapshots of the data accessible, and do the announcements in batches.

        :returns: Configuration for the announcement of new data to screen reader users.
        :rtype: :class:`AnnounceNewData` or :obj:`None <python:None>`
        """
        return self._announce_new_data

    @announce_new_data.setter
    @class_sensitive(AnnounceNewData)
    def announce_new_data(self, value):
        self._announce_new_data = value

    @property
    def custom_components(self) -> Optional[CustomAccessibilityComponents]:
        """Property which supports the definition of custom components added to the
        accessibility module. Defaults to :obj:`None <python:None>`.

        .. tip::

          Remember to add the component to the :meth:`keyboard_navigation.order`
          for the keyboard navigation to be usable.

        :returns: Custom components that have been added to the accessibility module.
        :rtype: :class:`CustomAccessibilityComponents` or :obj:`None <python:None>`

        """
        return self._custom_components

    @custom_components.setter
    @class_sensitive(CustomAccessibilityComponents)
    def custom_components(self, value):
        self._custom_components = value

    @property
    def description(self) -> Optional[str]:
        """A text description of the chart.

        If the Accessibility module is loaded, this option is included by default as a
        long description of the chart in the hidden screen reader information region.

        .. warning::

          Prefer using :meth:`Accessibility.linked_description` and
          :meth:`Options.caption` instead.

          Since Highcharts now supports captions and linked descriptions, it is preferred
          to define the description using those methods, as a visible caption/description
          benefits all users. If the ``accessibility.description`` option is defined, the
          linked description is ignored, and the caption is hidden from screen reader
          users.

        :returns: The description of the chart or :obj:`None <python:None>`
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._description

    @description.setter
    def description(self, value):
        value = validators.string(value, allow_empty = True)
        self._description = value

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enables accessibility functionality for the chart. Defaults to
        ``True``.

        .. seealso::

          For more information on how to include these features, and why this is
          recommended, see
          `Highcharts Accessibility <https://www.highcharts.com/docs/accessibility/accessibility-module>`_

        .. note::

          Highcharts will by default emit a warning to the console if the accessibility
          module is not loaded. Setting this option to false will override and silence the
          warning.

          Once the module is loaded, setting this option to false will disable the module
          for this chart.

        :returns: Flag indicating whether the :class:`Accessibility` module is enabled
          for the chart.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def high_contrast_theme(self) -> Any:
        """Theme to apply to the chart when Windows High Contrast Mode is detected.

        By default, a high contrast theme matching the high contrast system system colors
        is used.

        :returns: Theme to apply to the chart when high contrast mode is detected.
        :rtype: :class:`HighContrastTheme` or :obj:`None <python:None>`
        """
        return self._high_contrast_theme

    @high_contrast_theme.setter
    def high_contrast_theme(self, value):
        self._high_contrast_theme = value

    @property
    def keyboard_navigation(self) -> Optional[KeyboardNavigation]:
        """Options for keyboard navigation.

        :returns: Configuration for keyboard navigation.
        :rtype: :class:`KeyboardNavigation` or :obj:`None <python:None>`
        """
        return self._keyboard_navigation

    @keyboard_navigation.setter
    @class_sensitive(KeyboardNavigation)
    def keyboard_navigation(self, value):
        self._keyboard_navigation = value

    @property
    def landmark_verbosity(self) -> Optional[str]:
        """Amount of landmarks/regions to create for screen reader users.

        More landmarks can make navigation with screen readers easier, but can be
        distracting if there are lots of charts on the page.

        Three modes are available:

          * ``'all'``: Adds regions for all series, legend, information region.
          * ``'one'``: Adds a single landmark per chart.
          * ``'disabled'``: No landmarks are added.

        Defaults to ``'all'``.

        :returns: The landmark verbosity for screen readers to use.
        :rtype: :class:`str` or :obj:`None <python:None>`
        """
        return self._landmark_verbosity

    @landmark_verbosity.setter
    def landmark_verbosity(self, value):
        if not value:
            self._landmark_verbosity = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in constants.LANDMARK_VERBOSITY_VALUES:
                raise errors.HighchartsValueError(f'landmark_verbosity must be "all", '
                                                  f'"one", or "disabled". Was: {value}')
            self._landmark_verbosity = value

    @property
    def linked_description(self) -> Optional[str]:
        """Link the chart to an HTML element describing the contents of the chart.

        .. hint::

          It is always recommended to describe charts using visible text, to improve SEO
          as well as accessibility for users with disabilities.

          This option lets an HTML element with a description be linked to the chart, so
          that screen reader users can connect the two.

        By setting this option to a string, Highcharts runs the string as an HTML selector
        query on the entire document. If there is only a single match, this element is
        linked to the chart. The content of the linked element will be included in the
        chart description for screen reader users.

        By default, the chart looks for an adjacent sibling element with the
        ``highcharts-description`` (CSS) class.

        The feature can be disabled by setting the option to an empty string, or
        overridden by providing the :meth`Accessibility.description` option.
        Alternatively, the HTML element to link can be passed in directly as an HTML node.

        .. hint::

          If you need the description to be part of the exported image, consider using the
          caption feature.

          If you need the description to be hidden visually, use the
          :meth:`Accessibility.description` option.

        Defaults to ``'*[data-highcharts-chart="{index}"] + .highcharts-description'``.

        :returns: The CSS selector used to indicate the HTML element containing a
          description of the chart.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`

        """
        return self._linked_description

    @linked_description.setter
    def linked_description(self, value):
        self._linked_description = validators.string(value, allow_empty = True)

    @property
    def point(self) -> Optional[AccessibilityPoint]:
        """Options for describing individual points.

        :returns: Configuration for how to describe individual points on the chart.
        :rtype: :class:`AccessibilityPoint` or :obj:`None <python:None>`
        """
        return self._point

    @point.setter
    @class_sensitive(AccessibilityPoint)
    def point(self, value):
        self._point = value

    @property
    def screen_reader_section(self) -> Optional[ScreenReaderSection]:
        """Accessibility options for the screen reader information sections added before
        and after the chart.

        :returns: Configuration for the screen reader sections before and after the chart.
        :rtype: :class:`ScreenReaderSection` or :obj:`None <python:None>`
        """
        return self._screen_reader_section

    @screen_reader_section.setter
    @class_sensitive(ScreenReaderSection)
    def screen_reader_section(self, value):
        self._screen_reader_section = value

    @property
    def series(self) -> Optional[SeriesAccessibility]:
        """Accessibility options global to all data series.

        .. hint::

          Individual series can also have specific accessibility options set.

        :returns: Global accessibility options applied to all data series.
        :rtype: :class:`SeriesAccessibility` or :obj:`None <python:None>`
        """
        return self._series

    @series.setter
    @class_sensitive(SeriesAccessibility)
    def series(self, value):
        self._series = value

    @property
    def type_description(self) -> Optional[str]:
        """A text description of the chart type.

        If the Accessibility module is loaded, this will be included in the description
        of the chart in the screen reader information region.

        .. note::

          Highcharts will by default attempt to guess the chart type, but for more complex
          charts it is recommended to specify this property for clarity.

        :returns: A description fo the chart type.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._type_description

    @type_description.setter
    def type_description(self, value):
        self._type_description = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'announce_new_data': as_dict.get('announceNewData', None),
            'custom_components': as_dict.get('customComponents', None),
            'description': as_dict.get('description', None),
            'enabled': as_dict.get('enabled', None),
            'high_contrast_theme': as_dict.get('highContrastTheme', None),
            'keyboard_navigation': as_dict.get('keyboardNavigation', None),
            'landmark_verbosity': as_dict.get('landmarkVerbosity', None),
            'linked_description': as_dict.get('linkedDescription', None),
            'point': as_dict.get('point', None),
            'screen_reader_section': as_dict.get('screenReaderSection', None),
            'series': as_dict.get('series', None),
            'type_description': as_dict.get('typeDescription', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'announceNewData': self.announce_new_data,
            'customComponents': self.custom_components,
            'description': self.description,
            'enabled': self.enabled,
            'highContrastTheme': self.high_contrast_theme,
            'keyboardNavigation': self.keyboard_navigation,
            'landmarkVerbosity': self.landmark_verbosity,
            'linkedDescription': self.linked_description,
            'point': self.point,
            'screenReaderSection': self.screen_reader_section,
            'series': self.series,
            'typeDescription': self.type_description
        }

        return untrimmed


__all__ = [
    'Accessibility',
    'AnnounceNewData',
    'HighContrastTheme',
    'KeyboardNavigation',
    'AccessibilityPoint',
    'ScreenReaderSection',
    'SeriesAccessibility'
]



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.utility_classes.data_grouping import DataGroupingOptions as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'anchor': 'start',
      'approximation': 'windbarb',
      'date_time_label_formats': {
        'day': 'test',
        'hour': 'test',
        'millisecond': 'test',
        'minute': 'test',
        'month': 'test',
        'second': 'test',
        'week': 'test',
        'year': 'test'
      },
      'enabled': True,
      'first_anchor': 'start',
      'forced': True,
      'group_all': True,
      'group_pixel_width': 10,
      'last_anchor': 'end',
      'units': [
          [
              'millisecond',
              [1, 2, 5, 10, 20, 25, 50, 100, 200, 500]
          ],
          [
              'second',
              [1, 2, 5, 10, 15, 30]
          ],
          [
              'minute',
              [1, 2, 5, 10, 15, 30]
          ],
          [
              'hour',
              [1, 2, 3, 4, 6, 8, 12]
          ],
          [
              'day',
              [1]
          ],
          [
              'week',
              [1]
          ],
          [
              'month',
              [1, 3, 6]
          ],
          [
              'year',
              None
          ]
      ]
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/data_grouping/01.js', False, None),

    ('utility_classes/data_grouping/error-01.js', False, (errors.HighchartsValueError,
                                                          errors.HighchartsParseError,
                                                          JSONDecodeError,
                                                          TypeError,
                                                          ValueError)),

    ('utility_classes/data_grouping/01.js', True, None),

    ('utility_classes/data_grouping/error-01.js', True, (errors.HighchartsValueError,
                                                         errors.HighchartsParseError,
                                                         JSONDecodeError,
                                                         TypeError,
                                                         ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern


class DataClass(HighchartsMeta):
    """Definition of ranges for use in a choropleth map."""

    def __init__(self, **kwargs):
        self._color = None
        self._from_ = None
        self._name = None
        self._to = None

        self.color = kwargs.get('color', None)
        self.from_ = kwargs.get('from_', None)
        self.name = kwargs.get('name', None)
        self.to = kwargs.get('to', None)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the data class. Defaults to :obj:`None <python:None>`.

        If :obj:`None <python:None>`, the color is pulled from the global or
        chart-specific colors array.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def from_(self) -> Optional[int | float | Decimal]:
        """The point value at which the data class' value range begins. Defaults to
        :obj:`None <python:None>`.

        .. note::

          The range of each :class:`DataClass` is closed at both ends, but can be
          overridden by a subsequent :class:`DataClass` instance included in the
          :meth:`ColorAxis.data_classes` collection.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._from_

    @from_.setter
    def from_(self, value):
        self._from_ = validators.numeric(value, allow_empty = True)

    @property
    def name(self) -> Optional[str]:
        """The name of the data class as it should apear in the legend. Defaults to
        :obj:`None <python:None>`.

        If :obj:`None <python:None>`, it is automatically created based on the
        :meth:`from_ <DataClass.from_>` and :meth:`to <DataClass.to>` values.

        .. hint::

          For full programmatic control, :meth:`Legend.label_formatter` can be used. In
          the formatter (JavaScript) function, ``this.from`` and ``this.to`` can be
          accessed.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = validators.string(value, allow_empty = True)

    @property
    def to(self) -> Optional[int | float | Decimal]:
        """The point value at which the data class' value range ends. Defaults to
        :obj:`None <python:None>`.

        .. note::

          The range of each :class:`DataClass` is closed at both ends, but can be
          overridden by a subsequent :class:`DataClass` instance included in the
          :meth:`ColorAxis.data_classes` collection.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._to

    @to.setter
    def to(self, value):
        self._to = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'color': as_dict.get('color', None),
            'from_': as_dict.get('from', None),
            'name': as_dict.get('name', None),
            'to': as_dict.get('to', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'color': self.color,
            'from': self.from_,
            'name': self.name,
            'to': self.to
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

import datetime

from validator_collection import validators, checkers

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.series.data.cartesian import CartesianData
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.partial_fill import PartialFillOptions


class BarData(CartesianData):
    """Variant of :class:`CartesianData` which is used for data points in a column or bar
    graph context."""

    def __init__(self, **kwargs):
        self._border_color = None
        self._border_width = None
        self._dash_style = None
        self._point_width = None

        self.border_color = kwargs.get('border_color', None)
        self.border_width = kwargs.get('border_width', None)
        self.dash_style = kwargs.get('dash_style', None)
        self.point_width = kwargs.get('point_width', None)

        super().__init__(**kwargs)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the border surrounding each column or bar. Defaults to
        :obj:`None <python:None>`

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The width of the border surrounding each column or bar. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def dash_style(self) -> Optional[str]:
        """Name of the dash style to use for bar or column. Defaults to
        :obj:`None <python:None>`.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._dash_style

    @dash_style.setter
    def dash_style(self, value):
        if not value:
            self._dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'dash_style expects a recognized value'
                                                  f', but received: {value}')
            self._dash_style = value

    @property
    def point_width(self) -> Optional[int | float | Decimal]:
        """A pixel value specifying a fixed width for the column or bar. Defaults to
        :obj:`None <python:None>`.

        .. note::

          If specified, overrides the :meth:`point_width <BaseBarOptions.point_width>`
          provided for the series as a whole.

        .. note::

          The ``point_width`` affects the dimension that is *not* based on the point
          value.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._point_width

    @point_width.setter
    def point_width(self, value):
        self._point_width = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelRank',
                                      None) or as_dict.get('labelrank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),
            'marker': as_dict.get('marker', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),

            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'dash_style': as_dict.get('dashStyle', None),
            'point_width': as_dict.get('pointWidth', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'borderColor': self.border_color,
            'borderWidth': self.border_width,
            'dashStyle': self.dash_style,
            'pointWidth': self.point_width,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class WaterfallData(CartesianData):
    """Variant of :class:`CartesianData` which is used for data points in a waterfall
    chart."""

    def __init__(self, **kwargs):
        self._is_intermediate_sum = None
        self._is_sum = None

        self.is_intermediate_sum = kwargs.get('is_intermediate_sum', None)
        self.is_sum = kwargs.get('is_sum', None)

        super().__init__(**kwargs)

    @property
    def is_intermediate_sum(self) -> Optional[bool]:
        """When ``True``, the point acts as a summary column for the values added or
        subtracted since the last intermediate sum, or since the start of the series.
        Defaults to :obj:`None <python:None>`, which is interpreted as ``False``.

        .. warning::

          The ``y`` value is ignored.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._is_intermediate_sum

    @is_intermediate_sum.setter
    def is_intermediate_sum(self, value):
        if value is None:
            self._is_intermediate_sum = None
        else:
            self._is_intermediate_sum = bool(value)

    @property
    def is_sum(self) -> Optional[bool]:
        """When ``True``, the point displays the total sum across the entire series.
        Defaults to :obj:`None <python:None>`, which is interpreted as ``False``.

        .. warning::

          The ``y`` value is ignored.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._is_sum

    @is_sum.setter
    def is_sum(self, value):
        if value is None:
            self._is_sum = None
        else:
            self._is_sum = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelRank',
                                      None) or as_dict.get('labelrank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),
            'marker': as_dict.get('marker', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),

            'is_intermediate_sum': as_dict.get('isIntermediateSum', None),
            'is_sum': as_dict.get('isSum', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'isIntermediateSum': self.is_intermediate_sum,
            'isSum': self.is_sum,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class WindBarbData(CartesianData):
    """Variant of :class:`CartesianData` which is used for data points in a windbarb
    chart."""

    def __init__(self, **kwargs):
        self._direction = None
        self._value = None

        self.direction = kwargs.get('direction', None)
        self.value = kwargs.get('value', None)

        super().__init__(**kwargs)

    @property
    def direction(self) -> Optional[int | float | Decimal]:
        """The windirection in degrees, where 0 is north (pointing towards south).
        Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._direction

    @direction.setter
    def direction(self, value):
        self._direction = validators.numeric(value, allow_empty = True)

    @property
    def value(self) -> Optional[int | float | Decimal]:
        """The wind speed in meters per second. Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._value

    @value.setter
    def value(self, value_):
        self._value = validators.numeric(value_, allow_empty = True)

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'WindBarbData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls(x = None,
                             value = None,
                             direction = None,
                             y = None)
            elif checkers.is_iterable(item):
                if len(item) == 4:
                    as_dict = {
                        'x': item[0],
                        'value': item[1],
                        'direction': item[2],
                        'y': item[3]
                    }
                elif len(item) == 3:
                    as_dict = {
                        'x': item[0],
                        'value': item[1],
                        'direction': item[2]
                    }
                else:
                    raise errors.HighchartsValueError(f'data expects either a 4D or 3D '
                                                      f'collection. Collection received '
                                                      f'had {len(item)} dimensions.')

                as_obj = cls.from_dict(as_dict)
                if checkers.is_string(as_obj.x):
                    as_obj.name = as_obj.x
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a WindBarb Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')

            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelRank',
                                      None) or as_dict.get('labelrank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),
            'marker': as_dict.get('marker', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),

            'direction': as_dict.get('direction', None),
            'value': as_dict.get('value', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'direction': self.direction,
            'value': self.value,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class XRangeData(CartesianData):
    """Variant of :class:`CartesianData` which is used for data points in an X-Range
    series."""

    def __init__(self, **kwargs):
        self._partial_fill = None
        self._x2 = None

        self.partial_fill = kwargs.get('partial_fill', None)
        self.x2 = kwargs.get('x2', None)

        super().__init__(**kwargs)

    @property
    def partial_fill(self) -> Optional[PartialFillOptions]:
        """A partial fill for the data point, typically used to visualize how much of a
        task is performed. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`PartialFillOptions` or :obj:`None <python:None>`
        """
        return self._partial_fill

    @partial_fill.setter
    @class_sensitive(PartialFillOptions)
    def partial_fill(self, value):
        self._partial_fill = value

    @property
    def x(self) -> Optional[str | datetime.date | datetime.datetime | int | float | Decimal]:
        """The starting X-value of the range point. Defaults to :obj:`None <python:None>`.

        If :obj:`None <python:None>`, the point's position on the x-axis will be
        automatically determined based on its position in the series'
        :meth:`data <SeriesBase.data>` array. The first point will be given an ``x`` value
        of ``0``, or the series' :meth:`point_start <SeriesBase.point_start>` value.
        Each subsequent point will be incremented either by ``1`` or the value of
        :meth:`point_interval <SeriesBase.point_interval>`.

        :rtype: numeric or :class:`str <python:str>` or
          :class:`date <python:datetime.date>` or
          :class:`datetime <python:datetime.datetime>` or :obj:`None <python:None>`
        """
        return self._x

    @x.setter
    def x(self, value):
        if value is None:
            self._x = None
        else:
            if checkers.is_datetime(value):
                value = validators.datetime(value)
            elif checkers.is_date(value):
                value = validators.date(value)
            elif checkers.is_numeric(value):
                value = validators.numeric(value)
            else:
                value = validators.string(value)

            self._x = value

    @property
    def x2(self) -> Optional[str | datetime.date | datetime.datetime | int | float | Decimal]:
        """The ending X-value of the range point. Defaults to :obj:`None <python:None>`.

        If :obj:`None <python:None>`, the point's position on the x-axis will be
        automatically determined based on its position in the series'
        :meth:`data <SeriesBase.data>` array. The first point will be given an ``x2`` value
        of ``0``, or the series' :meth:`point_start <SeriesBase.point_start>` value.
        Each subsequent point will be incremented either by ``1`` or the value of
        :meth:`point_interval <SeriesBase.point_interval>`.

        :rtype: numeric or :class:`str <python:str>` or
          :class:`date <python:datetime.date>` or
          :class:`datetime <python:datetime.datetime>` or :obj:`None <python:None>`
        """
        return self._x2

    @x2.setter
    def x2(self, value):
        if value is None:
            self._x2 = None
        else:
            if checkers.is_datetime(value):
                value = validators.datetime(value)
            elif checkers.is_date(value):
                value = validators.date(value)
            elif checkers.is_numeric(value):
                value = validators.numeric(value)
            else:
                value = validators.string(value)

            self._x2 = value

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'XRangeData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = None
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be an X-Range Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelRank',
                                      None) or as_dict.get('labelrank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),
            'marker': as_dict.get('marker', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),

            'partial_fill': as_dict.get('partialFill', None),
            'x2': as_dict.get('x2', None),

        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'partialFill': self.partial_fill,
            'x': self.x,
            'x2': self.x2,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.options.accessibility.point import AccessibilityPoint


class SeriesKeyboardNavigation(HighchartsMeta):
    """Keyboard navigation support for the series."""

    def __init__(self, **kwargs):
        self._enabled = None

        self.enabled = kwargs.get('enabled', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions.series.accessibility.keyboardNavigation'

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enable accessibility functionality for the series.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'enabled': as_dict.get('enabled', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'enabled': self.enabled
        }

        return untrimmed


class TypeOptionsAccessibility(HighchartsMeta):
    """Accessibility options for a series."""

    def __init__(self, **kwargs):
        self._description = None
        self._description_format = None
        self._enabled = None
        self._expose_as_group_only = None
        self._keyboard_navigation = None
        self._point = None

        self.description = kwargs.get('description', None)
        self.description_format = kwargs.get('description_format', None)
        self.enabled = kwargs.get('enabled', None)
        self.expose_as_group_only = kwargs.get('expose_as_group_only', None)
        self.keyboard_navigation = kwargs.get('keyboard_navigation', None)
        self.point = kwargs.get('point', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions.series.accessibility'

    @property
    def description(self) -> Optional[str]:
        """Provide a description of the series, announced to screen readers.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._description

    @description.setter
    def description(self, value):
        self._description = validators.string(value, allow_empty = True)

    @property
    def description_format(self) -> Optional[str]:
        """Format to use for describing the data series group to assistive technology -
        including screen readers.

        Defaults to ``'{seriesDescription}{authorDescription}{axisDescription}'``.

        The series context and its subproperties are available under the variable
        ``{{series}}``, for example ``{{series.name}}`` for the series name, and
        ``{{series.points.length}}`` for the number of data points.

        The chart context and its subproperties are available under the variable
        ``{{chart}}``, for example ``{{chart.series.length}}`` for the number of series in
        the chart.

        ``{{seriesDescription}}`` refers to the automatic description of the series type
        and number of points added by Highcharts by default.

        ``{{authorDescription}}`` refers to the description added in
        ``series.description`` if one is present.

        ``{{axisDescription}}`` refers to the description added if the chart has multiple
        X or Y axes.

        :returns: Format string that applies to the description produced for the data
          series.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._description_format

    @description_format.setter
    def description_format(self, value):
        self._description_format = validators.string(value, allow_empty = True)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enable accessibility functionality for the series.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def expose_as_group_only(self) -> Optional[bool]:
        """If ``True``, expose only the series element to screen readers, not its points.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._expose_as_group_only

    @expose_as_group_only.setter
    def expose_as_group_only(self, value):
        if value is None:
            self._expose_as_group_only = None
        else:
            self._expose_as_group_only = bool(value)

    @property
    def keyboard_navigation(self) -> Optional[SeriesKeyboardNavigation]:
        """Keyboard navigation support for the series.

        :rtype: :class:`SeriesKeyboardNavigation` or :obj:`None <python:None>`
        """
        return self._keyboard_navigation

    @keyboard_navigation.setter
    @class_sensitive(SeriesKeyboardNavigation)
    def keyboard_navigation(self, value):
        self._keyboard_navigation = value

    @property
    def point(self) -> Optional[AccessibilityPoint]:
        """Point accessibility options for the series.

        :rtype: :class:`SeriesPoint` or :obj:`None <python:None>`
        """
        return self._point

    @point.setter
    @class_sensitive(AccessibilityPoint)
    def point(self, value):
        self._point = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'description': as_dict.get('description', None),
            'description_format': as_dict.get('description_format', None),
            'enabled': as_dict.get('enabled', None),
            'expose_as_group_only': as_dict.get('exposeAsGroupOnly', None),
            'keyboard_navigation': as_dict.get('keyboardNavigation', None),
            'point': as_dict.get('point', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'description': self.description,
            'descriptionFormat': self.description_format,
            'enabled': self.enabled,
            'exposeAsGroupOnly': self.expose_as_group_only,
            'keyboardNavigation': self.keyboard_navigation,
            'point': self.point
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.utility_classes.breadcrumbs import BreadcrumbOptions as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'button_spacing': 6,
      'button_theme': {
          'fill': '#fff'
      },
      'events': {
        'click': """function(event) { return true; }"""
      },
      'floating': True,
      'format': 'some format string',
      'formatter': """function () { return true; }""",
      'position': None,
      'relative_to': 'plot',
      'rtl': False,
      'separator': {
          'text': '>',
          'style': {
              'some-key': 'some-value'
          }
      },
      'use_html': False,
      'z_index': 3
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/breadcrumbs/01.js', False, None),

    ('utility_classes/breadcrumbs/error-01.js', False, (errors.HighchartsValueError,
                                                        errors.HighchartsParseError,
                                                        JSONDecodeError,
                                                        TypeError,
                                                        ValueError)),

    ('utility_classes/breadcrumbs/01.js', True, None),

    ('utility_classes/breadcrumbs/error-01.js', True, (errors.HighchartsValueError,
                                                       errors.HighchartsParseError,
                                                       JSONDecodeError,
                                                       TypeError,
                                                       ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.axes.parallel_axes import ParallelAxesOptions as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    # Y-Axis propeties only
    ({
      'max_color': '#ccc',
      'min_color': '#000',
      'stack_labels': {
        'align': 'center',
        'allowOverlap': True,
        'animation': {
            'defer': 5
        },
        'backgroundColor': {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.234,
                'y1': 0.345,
                'y2': 0.456
            },
            'stops': [
                [0.12, '#999'],
                [0.34, '#fff']
            ]
        },
        'borderColor': '#999999',
        'borderRadius': 24,
        'borderWidth': 1,
        'className': 'some-class-name',
        'color': '#000000',
        'crop': True,
        'defer': False,
        'enabled': True,
        'filter': {
            'operator': '>=',
            'property': 'some_property',
            'value': 123
        },
        'format': 'some format',
        'formatter': """function() { return true; }""",
        'inside': True,
        'nullFormat': 'some format',
        'nullFormatter': """function() { return true; }""",
        'overflow': 'none',
        'padding': 12,
        'position': 'center',
        'rotation': 0,
        'shadow': False,
        'shape': 'rect',
        'style': 'style goes here',
        'useHTML': False,
        'verticalAlign': 'top',
        'x': 10,
        'y': 20,
        'z': 0
      },
      'stops': [
          [0.0, '#ccc'],
          [0.1, '#fff'],
          [0.25, '#999'],
          [1.0, '#ff0000']
      ],
      'tooltip_value_format': 'format string goes here'
    }, None),
    # XAxis properties only
    ({
      'max_color': '#ccc',
      'min_color': '#000',
      'stack_labels': {
        'align': 'center',
        'allowOverlap': True,
        'animation': {
            'defer': 5
        },
        'backgroundColor': {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.234,
                'y1': 0.345,
                'y2': 0.456
            },
            'stops': [
                [0.12, '#999'],
                [0.34, '#fff']
            ]
        },
        'borderColor': '#999999',
        'borderRadius': 24,
        'borderWidth': 1,
        'className': 'some-class-name',
        'color': '#000000',
        'crop': True,
        'defer': False,
        'enabled': True,
        'filter': {
            'operator': '>=',
            'property': 'some_property',
            'value': 123
        },
        'format': 'some format',
        'formatter': """function() { return true; }""",
        'inside': True,
        'nullFormat': 'some format',
        'nullFormatter': """function() { return true; }""",
        'overflow': 'none',
        'padding': 12,
        'position': 'center',
        'rotation': 0,
        'shadow': False,
        'shape': 'rect',
        'style': 'style goes here',
        'useHTML': False,
        'verticalAlign': 'top',
        'x': 10,
        'y': 20,
        'z': 0
      },
      'stops': [
          [0.0, '#ccc'],
          [0.1, '#fff'],
          [0.25, '#999'],
          [1.0, '#ff0000']
      ],
      'tooltip_value_format': 'format string goes here',

      'crosshair': {
          'className': 'some-class-name',
          'color': '#ccc',
          'dashStyle': 'Dash',
          'snap': True,
          'width': 40,
          'zIndex': 6
      },
      'height': 60,
      'left': 5,
      'line_color': '#999',
      'line_width': 1,
      'show_empty': True,
      'top': 340,
      'width': 300,
    }, None),
    # with GenericAxis and NumericAxis properties
    ({
      'max_color': '#ccc',
      'min_color': '#000',
      'stack_labels': {
        'align': 'center',
        'allowOverlap': True,
        'animation': {
            'defer': 5
        },
        'backgroundColor': {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.234,
                'y1': 0.345,
                'y2': 0.456
            },
            'stops': [
                [0.12, '#999'],
                [0.34, '#fff']
            ]
        },
        'borderColor': '#999999',
        'borderRadius': 24,
        'borderWidth': 1,
        'className': 'some-class-name',
        'color': '#000000',
        'crop': True,
        'defer': False,
        'enabled': True,
        'filter': {
            'operator': '>=',
            'property': 'some_property',
            'value': 123
        },
        'format': 'some format',
        'formatter': """function() { return true; }""",
        'inside': True,
        'nullFormat': 'some format',
        'nullFormatter': """function() { return true; }""",
        'overflow': 'none',
        'padding': 12,
        'position': 'center',
        'rotation': 0,
        'shadow': False,
        'shape': 'rect',
        'style': 'style goes here',
        'useHTML': False,
        'verticalAlign': 'top',
        'x': 10,
        'y': 20,
        'z': 0
      },
      'stops': [
          [0.0, '#ccc'],
          [0.1, '#fff'],
          [0.25, '#999'],
          [1.0, '#ff0000']
      ],
      'tooltip_value_format': 'format string goes here',

      'crosshair': {
          'className': 'some-class-name',
          'color': '#ccc',
          'dashStyle': 'Dash',
          'snap': True,
          'width': 40,
          'zIndex': 6
      },
      'height': 60,
      'left': 5,
      'line_color': '#999',
      'line_width': 1,
      'show_empty': True,
      'top': 340,
      'width': 300,

      'accessibility': {
          'description': 'Description goes here',
          'enabled': True,
          'rangeDescription': 'Range description goes here'
      },
      'align_ticks': False,
      'allow_decimals': True,
      'alternate_grid_color': '#ccc',
      'angle': 15,
      'breaks': [
          {
           'breakSize': 20,
           'from': 0,
           'repeat': 1,
           'to': 100
          },
          {
           'breakSize': 50,
           'from': 100,
           'repeat': 1,
           'to': 1000
          }
      ],
      'categories': [
          'Category 1',
          'Category 2',
          'Category 3',
          'Category 4'
      ],
      'ceiling': 120,
      'class_name': 'some-class-name',
      'date_time_label_formats': {
        'day': 'test',
        'hour': 'test',
        'millisecond': 'test',
        'minute': 'test',
        'month': 'test',
        'second': 'test',
        'week': 'test',
        'year': 'test'
      },
      'end_on_tick': False,
      'events': {
        'afterBreaks': """function(event) { return true; }""",
        'afterSetExtremes': """function(event) { return true; }""",
        'pointBreak': """function(event) { return true; }""",
        'setExtremes': """function(event) { return true; }"""
      },
      'floor': 0,
      'grid_line_color': '#ccc',
      'grid_line_dash_style': 'Solid',
      'grid_line_interpolation': 'circle',
      'grid_line_width': 1,
      'grid_z_index': 3,
      'id': 'some-id',
      'labels': {
          'align': 'center',
          'allowOverlap': False,
          'autoRotation': [-45],
          'autoRotationLimit': 80,
          'distance': 12,
          'enabled': True,
          'format': 'some format string',
          'formatter': """function () { return true; }""",
          'overflow': 'allow',
          'padding': 12,
          'position3d': 'offset',
          'reserveSpace': True,
          'rotation': 24,
          'skew3d': False,
          'staggerLines': 0,
          'step': 2,
          'style': 'some-style-string',
          'useHTML': False,
          'x': 5,
          'y': -10,
          'zIndex': 6
      },
      'linked_to': 3,
      'margin': 12,
      'max': 1000,
      'max_padding': 12,
      'min': 0,
      'min_range': 5,
      'min_tick_interval': 1,
      'minor_grid_line_color': '#999',
      'minor_grid_line_dash_style': 'Dash',
      'minor_grid_line_width': 1,
      'minor_tick_color': '#ccc',
      'minor_tick_interval': 0.1,
      'minor_tick_position': 'outside',
      'minor_ticks': True,
      'minor_tick_width': 1,
      'min_padding': 8,
      'offset': 0,
      'opposite': False,
      'pane': 1,
      'panning_enabled': True,
      'reversed': False,
      'reversed_stacks': False,
      'show_first_label': True,
      'show_last_label': True,
      'soft_max': 10,
      'soft_min': 6,
      'start_of_week': 1,
      'start_on_tick': False,
      'tick_amount': 5,
      'tick_color': '#000',
      'tick_interval': 5,
      'tick_length': 8,
      'tickmark_placement': 'on',
      'tick_pixel_interval': 8,
      'tick_position': 'outside',
      'tick_positioner': """function() { return true; }""",
      'tick_width': 1,
      'title': {
          'align': 'low',
          'margin': 20,
          'offset': 0,
          'position3d': 'offset',
          'reserveSpace': True,
          'rotation': 0,
          'skew3d': False,
          'style': 'some-style-string',
          'text': 'The Axis Title',
          'textAlign': 'center',
          'useHTML': False,
          'x': 5,
          'y': 10
      },
      'type': 'linear',
      'unique_names': True,
      'units': [
          [
              'millisecond',
              [1, 2, 5, 10, 20, 25, 50, 100, 200, 500]
          ],
          [
              'second',
              [1, 2, 5, 10, 15, 30]
          ],
          [
              'minute',
              [1, 2, 5, 10, 15, 30]
          ],
          [
              'hour',
              [1, 2, 3, 4, 6, 8, 12]
          ],
          [
              'day',
              [1]
          ],
          [
              'week',
              [1]
          ],
          [
              'month',
              [1, 3, 6]
          ],
          [
              'year',
              None
          ]
      ],
      'visible': True,
      'z_index': 3,
      'zoom_enabled': True

    }, None),

]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('axes/parallel_axes/01.js', False, None),
    ('axes/parallel_axes/02.js', False, None),

    ('axes/parallel_axes/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
    ('axes/parallel_axes/error-02.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('axes/parallel_axes/01.js', True, None),
    ('axes/parallel_axes/02.js', True, None),

    ('axes/parallel_axes/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
    ('axes/parallel_axes/error-02.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta


class SeriesTypeDescriptions(HighchartsMeta):
    """Descriptions of lesser known series types. The relevant description is added to
    the screen reader information region when these series types are used."""

    def __init__(self, **kwargs):
        self._arearange = None
        self._areasplinerange = None
        self._boxplot = None
        self._bubble = None
        self._columnrange = None
        self._errorbar = None
        self._funnel = None
        self._pyramid = None
        self._waterfall = None

        self.arearange = kwargs.get('arearange', None)
        self.areasplinerange = kwargs.get('areasplinerange', None)
        self.boxplot = kwargs.get('boxplot', None)
        self.bubble = kwargs.get('bubble', None)
        self.columnrange = kwargs.get('columnrange', None)
        self.errorbar = kwargs.get('errorbar', None)
        self.funnel = kwargs.get('funnel', None)
        self.pyramid = kwargs.get('pyramid', None)
        self.waterfall = kwargs.get('waterfall', None)

    @property
    def arearange(self) -> Optional[str]:
        """Defaults to ``'Arearange charts are line charts displaying a range between a lower and higher value for each point.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._arearange

    @arearange.setter
    def arearange(self, value):
        self._arearange = validators.string(value, allow_empty = True)

    @property
    def areasplinerange(self) -> Optional[str]:
        """Defaults to
        ``'These charts are line charts displaying a range between a lower and higher value for each point.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._areasplinerange

    @areasplinerange.setter
    def areasplinerange(self, value):
        self._areasplinerange = validators.string(value, allow_empty = True)

    @property
    def boxplot(self) -> Optional[str]:
        """Defaults to ``'Box plot charts are typically used to display groups of statistical data. Each data point in the chart can have up to 5 values: minimum, lower quartile, median, upper quartile, and maximum.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._boxplot

    @boxplot.setter
    def boxplot(self, value):
        self._boxplot = validators.string(value, allow_empty = True)

    @property
    def bubble(self) -> Optional[str]:
        """Defaults to ``'Bubble charts are scatter charts where each data point also has a size value.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._bubble

    @bubble.setter
    def bubble(self, value):
        self._bubble = validators.string(value, allow_empty = True)

    @property
    def columnrange(self) -> Optional[str]:
        """Defaults to
        ``'Columnrange charts are column charts displaying a range between a lower and higher value for each point.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._columnrange

    @columnrange.setter
    def columnrange(self, value):
        self._columnrange = validators.string(value, allow_empty = True)

    @property
    def errorbar(self) -> Optional[str]:
        """Defaults to ``'Errorbar series are used to display the variability of the data.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._errorbar

    @errorbar.setter
    def errorbar(self, value):
        self._errorbar = validators.string(value, allow_empty = True)

    @property
    def funnel(self) -> Optional[str]:
        """Defaults to ``'Funnel charts are used to display reduction of data in stages.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._funnel

    @funnel.setter
    def funnel(self, value):
        self._funnel = validators.string(value, allow_empty = True)

    @property
    def pyramid(self) -> Optional[str]:
        """Defaults to ``'Pyramid charts consist of a single pyramid with item heights corresponding to each point value.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._pyramid

    @pyramid.setter
    def pyramid(self, value):
        self._pyramid = validators.string(value, allow_empty = True)

    @property
    def waterfall(self) -> Optional[str]:
        """Defaults to ``'A waterfall chart is a column chart where each column contributes towards a total end value.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._waterfall

    @waterfall.setter
    def waterfall(self, value):
        self._waterfall = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'arearange': as_dict.get('arearange', None),
            'areasplinerange': as_dict.get('areasplinerange', None),
            'boxplot': as_dict.get('boxplot', None),
            'bubble': as_dict.get('bubble', None),
            'columnrange': as_dict.get('columnrange', None),
            'errorbar': as_dict.get('errorbar', None),
            'funnel': as_dict.get('funnel', None),
            'pyramid': as_dict.get('pyramid', None),
            'waterfall': as_dict.get('waterfall', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'arearange': self.arearange,
            'areasplinerange': self.areasplinerange,
            'boxplot': self.boxplot,
            'bubble': self.bubble,
            'columnrange': self.columnrange,
            'errorbar': self.errorbar,
            'funnel': self.funnel,
            'pyramid': self.pyramid,
            'waterfall': self.waterfall
        }

        return untrimmed


class SeriesSummaryLanguageOptions(HighchartsMeta):
    """Language configuration for the series main summary.

    Each series type has two modes:

      #. This series type is the only series type used in the chart
      #. This is a combination chart with multiple series types

    .. note::

      If a definition does not exist for the specific series type and mode, the
      ``'default'`` language definitions are used.

      Chart and its subproperties can be accessed with the ``{chart}`` variable. The
      series and its subproperties can be accessed with the ``{series}`` variable.

      The series index (starting from 1) can be accessed with the ``{seriesNumber}``
      variable.

    """

    def __init__(self, **kwargs):
        self._bar = None
        self._bar_combination = None
        self._boxplot = None
        self._boxplot_combination = None
        self._bubble = None
        self._bubble_combination = None
        self._column = None
        self._column_combination = None
        self._default = None
        self._default_combination = None
        self._line = None
        self._line_combination = None
        self._map = None
        self._map_combination = None
        self._mapbubble = None
        self._mapbubble_combination = None
        self._mapline = None
        self._mapline_combination = None
        self._pie = None
        self._pie_combination = None
        self._scatter = None
        self._scatter_combination = None
        self._spline = None
        self._spline_combination = None

        self.bar = kwargs.get('bar',
                              constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('bar'))
        self.bar_combination = kwargs.get('bar_combination',
                                          constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('bar_combination'))
        self.boxplot = kwargs.get('boxplot',
                                  constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('boxplot'))
        self.boxplot_combination = kwargs.get('boxplot_combination',
                                              constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('boxplot_combination'))
        self.bubble = kwargs.get('bubble',
                                 constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('bubble'))
        self.bubble_combination = kwargs.get('bubble_combination',
                                             constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('bubble_combination'))
        self.column = kwargs.get('column',
                                  constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('column'))
        self.column_combination = kwargs.get('column_combination',
                                             constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('column_combination'))
        self.default = kwargs.get('default',
                                   constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('default'))
        self.default_combination = kwargs.get('default_combination',
                                              constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('default_combination'))
        self.line = kwargs.get('line',
                                constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('line'))
        self.line_combination = kwargs.get('line_combination',
                                           constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('line_combination'))
        self.map = kwargs.get('map',
                               constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('map'))
        self.map_combination = kwargs.get('map_combination',
                                          constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('map_combination'))
        self.mapbubble = kwargs.get('mapbubble',
                                     constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('mapbubble'))
        self.mapbubble_combination = kwargs.get('mapbubble_combination',
                                                constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('mapbubble_combination'))
        self.mapline = kwargs.get('mapline',
                                   constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('mapline'))
        self.mapline_combination = kwargs.get('mapline_combination',
                                              constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('mapline_combination'))
        self.pie = kwargs.get('pie',
                               constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('pie'))
        self.pie_combination = kwargs.get('pie_combination',
                                          constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('pie_combination'))
        self.scatter = kwargs.get('scatter',
                                   constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('scatter'))
        self.scatter_combination = kwargs.get('scatter_combination',
                                              constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('scatter_combination'))
        self.spline = kwargs.get('spline',
                                  constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('spline'))
        self.spline_combination = kwargs.get('spline_combination',
                                             constants.DEFAULT_LANG_ACS_SERIES_SUMMARY.get('spline_combination'))

    @property
    def bar(self) -> Optional[str]:
        """Defaults to ``'{series.name}, bar series {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, bars, bar)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._bar

    @bar.setter
    def bar(self, value):
        self._bar = validators.string(value, allow_empty = True)

    @property
    def bar_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length}. Bar series with {series.points.length} {#plural(series.points.length, bars, bar)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._bar_combination

    @bar_combination.setter
    def bar_combination(self, value):
        self._bar_combination = validators.string(value, allow_empty = True)

    @property
    def boxplot(self) -> Optional[str]:
        """Defaults to ``'{series.name}, boxplot {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, boxes, box)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._boxplot

    @boxplot.setter
    def boxplot(self, value):
        self._boxplot = validators.string(value, allow_empty = True)

    @property
    def boxplot_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length}. Boxplot with {series.points.length} {#plural(series.points.length, boxes, box)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._boxplot_combination

    @boxplot_combination.setter
    def boxplot_combination(self, value):
        self._boxplot_combination = validators.string(value, allow_empty = True)

    @property
    def bubble(self) -> Optional[str]:
        """Defaults to ``'{series.name}, bubble series {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, bubbles, bubble)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._bubble

    @bubble.setter
    def bubble(self, value):
        self._bubble = validators.string(value, allow_empty = True)

    @property
    def bubble_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length}. Bubble series with {series.points.length} {#plural(series.points.length, bubbles, bubble)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._bubble_combination

    @bubble_combination.setter
    def bubble_combination(self, value):
        self._bubble_combination = validators.string(value, allow_empty = True)

    @property
    def column(self) -> Optional[str]:
        """Defaults to ``'{series.name}, bar series {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, bars, bar)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._column

    @column.setter
    def column(self, value):
        self._column = validators.string(value, allow_empty = True)

    @property
    def column_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length}. Bar series with {series.points.length} {#plural(series.points.length, bars, bar)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._column_combination

    @column_combination.setter
    def column_combination(self, value):
        self._column_combination = validators.string(value, allow_empty = True)

    @property
    def default(self) -> Optional[str]:
        """Defaults to ``'{series.name}, series {seriesNumber} of {chart.series.length} with {series.points.length} data {#plural(series.points.length, points, point)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._default

    @default.setter
    def default(self, value):
        self._default = validators.string(value, allow_empty = True)

    @property
    def default_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length} with {series.points.length} data {#plural(series.points.length, points, point)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._default_combination

    @default_combination.setter
    def default_combination(self, value):
        self._default_combination = validators.string(value, allow_empty = True)

    @property
    def line(self) -> Optional[str]:
        """Defaults to ``'{series.name}, line {seriesNumber} of {chart.series.length} with {series.points.length} data {#plural(series.points.length, points, point)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._line

    @line.setter
    def line(self, value):
        self._line = validators.string(value, allow_empty = True)

    @property
    def line_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length}. Line with {series.points.length} data {#plural(series.points.length, points, point)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._line_combination

    @line_combination.setter
    def line_combination(self, value):
        self._line_combination = validators.string(value, allow_empty = True)

    @property
    def map(self) -> Optional[str]:
        """Defaults to ``'{series.name}, map {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, areas, area)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._map

    @map.setter
    def map(self, value):
        self._map = validators.string(value, allow_empty = True)

    @property
    def map_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length}. Map with {series.points.length} {#plural(series.points.length, areas, area)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._map_combination

    @map_combination.setter
    def map_combination(self, value):
        self._map_combination = validators.string(value, allow_empty = True)

    @property
    def mapbubble(self) -> Optional[str]:
        """Defaults to ``'{series.name}, bubble series {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, bubbles, bubble)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._mapbubble

    @mapbubble.setter
    def mapbubble(self, value):
        self._mapbubble = validators.string(value, allow_empty = True)

    @property
    def mapbubble_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length}. Bubble series with {series.points.length} {#plural(series.points.length, bubbles, bubble)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._mapbubble_combination

    @mapbubble_combination.setter
    def mapbubble_combination(self, value):
        self._mapbubble_combination = validators.string(value, allow_empty = True)

    @property
    def mapline(self) -> Optional[str]:
        """Defaults to ``'{series.name}, line {seriesNumber} of {chart.series.length} with {series.points.length} data {#plural(series.points.length, points, point)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._mapline

    @mapline.setter
    def mapline(self, value):
        self._mapline = validators.string(value, allow_empty = True)

    @property
    def mapline_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length}. Line with {series.points.length} data {#plural(series.points.length, points, point)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._mapline_combination

    @mapline_combination.setter
    def mapline_combination(self, value):
        self._mapline_combination = validators.string(value, allow_empty = True)

    @property
    def pie(self) -> Optional[str]:
        """Defaults to ``'{series.name}, pie {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, slices, slice)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._pie

    @pie.setter
    def pie(self, value):
        self._pie = validators.string(value, allow_empty = True)

    @property
    def pie_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length}. Pie with {series.points.length} {#plural(series.points.length, slices, slice)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._pie_combination

    @pie_combination.setter
    def pie_combination(self, value):
        self._pie_combination = validators.string(value, allow_empty = True)

    @property
    def scatter(self) -> Optional[str]:
        """Defaults to ``'{series.name}, scatter plot {seriesNumber} of {chart.series.length} with {series.points.length} {#plural(series.points.length, points, point)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._scatter

    @scatter.setter
    def scatter(self, value):
        self._scatter = validators.string(value, allow_empty = True)

    @property
    def scatter_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length}, scatter plot with {series.points.length} {#plural(series.points.length, points, point)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._scatter_combination

    @scatter_combination.setter
    def scatter_combination(self, value):
        self._scatter_combination = validators.string(value, allow_empty = True)

    @property
    def spline(self) -> Optional[str]:
        """Defaults to ``'{series.name}, line {seriesNumber} of {chart.series.length} with {series.points.length} data {#plural(series.points.length, points, point)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._spline

    @spline.setter
    def spline(self, value):
        self._spline = validators.string(value, allow_empty = True)

    @property
    def spline_combination(self) -> Optional[str]:
        """Defaults to
        ``'{series.name}, series {seriesNumber} of {chart.series.length}. Line with {series.points.length} data {#plural(series.points.length, points, point)}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._spline_combination

    @spline_combination.setter
    def spline_combination(self, value):
        self._spline_combination = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'bar': as_dict.get('bar', None),
            'bar_combination': as_dict.get('barCombination', None),
            'boxplot': as_dict.get('boxplot', None),
            'boxplot_combination': as_dict.get('boxplotCombination', None),
            'bubble': as_dict.get('bubble', None),
            'bubble_combination': as_dict.get('bubbleCombination', None),
            'column': as_dict.get('column', None),
            'column_combination': as_dict.get('columnCombination', None),
            'default': as_dict.get('default', None),
            'default_combination': as_dict.get('defaultCombination', None),
            'line': as_dict.get('line', None),
            'line_combination': as_dict.get('lineCombination', None),
            'map': as_dict.get('map', None),
            'map_combination': as_dict.get('mapCombination', None),
            'mapbubble': as_dict.get('mapbubble', None),
            'mapbubble_combination': as_dict.get('mapbubbleCombination', None),
            'mapline': as_dict.get('mapline', None),
            'mapline_combination': as_dict.get('maplineCombination', None),
            'pie': as_dict.get('pie', None),
            'pie_combination': as_dict.get('pieCombination', None),
            'scatter': as_dict.get('scatter', None),
            'scatter_combination': as_dict.get('scatterCombination', None),
            'spline': as_dict.get('spline', None),
            'spline_combination': as_dict.get('splineCombination', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'bar': self.bar,
            'barCombination': self.bar_combination,
            'boxplot': self.boxplot,
            'boxplotCombination': self.boxplot_combination,
            'bubble': self.bubble,
            'bubbleCombination': self.bubble_combination,
            'column': self.column,
            'columnCombination': self.column_combination,
            'default': self.default,
            'defaultCombination': self.default_combination,
            'line': self.line,
            'lineCombination': self.line_combination,
            'map': self.map,
            'mapCombination': self.map_combination,
            'mapbubble': self.mapbubble,
            'mapbubbleCombination': self.mapbubble_combination,
            'mapline': self.mapline,
            'maplineCombination': self.mapline_combination,
            'pie': self.pie,
            'pieCombination': self.pie_combination,
            'scatter': self.scatter,
            'scatterCombination': self.scatter_combination,
            'spline': self.spline,
            'splineCombination': self.spline_combination
        }

        return untrimmed


class SeriesLanguageOptions(HighchartsMeta):
    """Language configuration for different series types.

    .. hint::

      For more dynamic control over the series element descriptions, see
      :meth:`Accessibility.series_description_formatter`.

    """

    def __init__(self, **kwargs):
        self._description = None
        self._null_point_value = None
        self._point_annotations_description = None
        self._summary = None
        self._x_axis_description = None
        self._y_axis_description = None

        self.description = kwargs.get('description', None)
        self.null_point_value = kwargs.get('null_point_value', None)
        self.point_annotations_description = kwargs.get('point_annotations_description',
                                                        None)
        self.summary = kwargs.get('summary', None)
        self.x_axis_description = kwargs.get('x_axis_description', None)
        self.y_axis_description = kwargs.get('y_axis_description', None)

    @property
    def description(self) -> Optional[str]:
        """User supplied description text. Defaults to:
        ``''``.

        .. note::

          This is added in the point comment description by default if present.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._description

    @description.setter
    def description(self, value):
        self._description = validators.string(value, allow_empty = True)

    @property
    def null_point_value(self) -> Optional[str]:
        """Description for the value of null points. Defaults to:
        ``''``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._null_point_value

    @null_point_value.setter
    def null_point_value(self, value):
        self._null_point_value = validators.string(value, allow_empty = True)

    @property
    def point_annotations_description(self) -> Optional[str]:
        """Description for annotations on a point, as it is made available to assistive
        technology. Defaults to:
        ``''``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._point_annotations_description

    @point_annotations_description.setter
    def point_annotations_description(self, value):
        self._point_annotations_description = validators.string(value, allow_empty = True)

    @property
    def summary(self) -> Optional[SeriesSummaryLanguageOptions]:
        """Language configuration for the series main summary.

        Each series type has two modes:

          #. This series type is the only series type used in the chart
          #. This is a combination chart with multiple series types

        .. note::

          If a definition does not exist for the specific series type and mode, the
          ``'default'`` language definitions are used.

          Chart and its subproperties can be accessed with the ``{chart}`` variable. The
          series and its subproperties can be accessed with the ``{series}`` variable.

          The series index (starting from 1) can be accessed with the ``{seriesNumber}``
          variable.

        :rtype: :class:`SeriesSummaryLanguageOptions` or :obj:`None <python:None>`
        """
        return self._summary

    @summary.setter
    @class_sensitive(SeriesSummaryLanguageOptions)
    def summary(self, value):
        self._summary = value

    @property
    def x_axis_description(self) -> Optional[str]:
        """xAxis description for series if there are multiple xAxes in the chart. Defaults
        to: ``''``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._x_axis_description

    @x_axis_description.setter
    def x_axis_description(self, value):
        self._x_axis_description = validators.string(value, allow_empty = True)

    @property
    def y_axis_description(self) -> Optional[str]:
        """yAxis description for series if there are multiple yAxes in the chart. Defaults
        to: ``''``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._y_axis_description

    @y_axis_description.setter
    def y_axis_description(self, value):
        self._y_axis_description = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'description': as_dict.get('description', None),
            'null_point_value': as_dict.get('nullPointValue', None),
            'point_annotations_description': as_dict.get('pointAnnotationsDescription',
                                                         None),
            'summary': as_dict.get('summary', None),
            'x_axis_description': as_dict.get('xAxisDescription', None),
            'y_axis_description': as_dict.get('yAxisDescription', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'description': self.description,
            'nullPointValue': self.null_point_value,
            'pointAnnotationsDescription': self.point_annotations_description,
            'summary': self.summary,
            'xAxisDescription': self.x_axis_description,
            'yAxisDescription': self.y_axis_description
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class DateTimeLabelFormats(HighchartsMeta):
    """For series on datetime axes, the date format in the tooltip's header will by default be
    guessed based on the closest data points. This class gives the default string
    representations used for each unit.
    """

    def __init__(self, **kwargs):
        self._day = None
        self._hour = None
        self._millisecond = None
        self._minute = None
        self._month = None
        self._second = None
        self._week = None
        self._year = None

        self.day = kwargs.get('day', None)
        self.hour = kwargs.get('hour', None)
        self.millisecond = kwargs.get('millisecond', None)
        self.minute = kwargs.get('minute', None)
        self.month = kwargs.get('month', None)
        self.second = kwargs.get('second', None)
        self.week = kwargs.get('week', None)
        self.year = kwargs.get('year', None)

    @property
    def day(self) -> Optional[str]:
        """Label format to apply to days. Defaults to
        ``'%A, %b %e, %Y'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._day

    @day.setter
    def day(self, value):
        self._day = validators.string(value, allow_empty = True)

    @property
    def hour(self) -> Optional[str]:
        """Label format to apply to hours. Defaults to
        ``'%A, %b %e, %H:%M'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._hour

    @hour.setter
    def hour(self, value):
        self._hour = validators.string(value, allow_empty = True)

    @property
    def millisecond(self) -> Optional[str]:
        """Label format to apply to milliseconds. Defaults to
        ``'%A, %b %e, %H:%M:%S.%L'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._millisecond

    @millisecond.setter
    def millisecond(self, value):
        self._millisecond = validators.string(value, allow_empty = True)

    @property
    def minute(self) -> Optional[str]:
        """Label format to apply to minutes. Defaults to
        ``'%A, %b %e, %H:%M'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._minute

    @minute.setter
    def minute(self, value):
        self._minute = validators.string(value, allow_empty = True)

    @property
    def month(self) -> Optional[str]:
        """Label format to apply to months. Defaults to
        ``'%B %Y'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._month

    @month.setter
    def month(self, value):
        self._month = validators.string(value, allow_empty = True)

    @property
    def second(self) -> Optional[str]:
        """Label format to apply to seconds. Defaults to
        ``'%A, %b %e, %H:%M:%S'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._second

    @second.setter
    def second(self, value):
        self._second = validators.string(value, allow_empty = True)

    @property
    def week(self) -> Optional[str]:
        """Label format to apply to weeks. Defaults to
        ``'Week from %A, %b %e, %Y'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._week

    @week.setter
    def week(self, value):
        self._week = validators.string(value, allow_empty = True)

    @property
    def year(self) -> Optional[str]:
        """Label format to apply to years. Defaults to
        ``'%Y'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._year

    @year.setter
    def year(self, value):
        self._year = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'day': as_dict.get('day', None),
            'hour': as_dict.get('hour', None),
            'millisecond': as_dict.get('millisecond', None),
            'minute': as_dict.get('minute', None),
            'month': as_dict.get('month', None),
            'second': as_dict.get('second', None),
            'week': as_dict.get('week', None),
            'year': as_dict.get('year', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'day': self.day,
            'hour': self.hour,
            'millisecond': self.millisecond,
            'minute': self.minute,
            'month': self.month,
            'second': self.second,
            'week': self.week,
            'year': self.year
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.drilldown import Drilldown as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'active_axis_label_style': {
          'cursor': 'style string',
          'color': '#999'
      },
      'active_data_label_style': {
          'cursor': 'style string',
          'color': '#999'
      },
      'allow_point_drilldown': True,
      'animation': False,
      'breadcrumbs': {
        'buttonSpacing': 6,
        'buttonTheme': {
            'fill': '#fff'
        },
        'events': {
          'click': """function(event) { return true; }"""
        },
        'floating': True,
        'format': 'some format string',
        'formatter': """function () { return true; }""",
        'position': None,
        'relativeTo': 'plot',
        'rtl': False,
        'separator': {
            'style': {
                'some-key': 'some-value'
            },
            'text': '>',
        },
        'useHTML': False,
        'zIndex': 3
      }
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('drilldown/01.js', False, None),

    ('drilldown/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('drilldown/01.js', True, None),

    ('drilldown/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors, constants
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern


class CrosshairOptions(HighchartsMeta):
    """Configuration options for a crosshair that follows either the mouse pointer or the
    hovered point."""

    def __init__(self, **kwargs):
        self._class_name = None
        self._color = None
        self._dash_style = None
        self._snap = None
        self._width = None
        self._z_index = None

        self.class_name = kwargs.get('class_name', None)
        self.color = kwargs.get('color', None)
        self.dash_style = kwargs.get('dash_style', None)
        self.snap = kwargs.get('snap', None)
        self.width = kwargs.get('width', None)
        self.z_index = kwargs.get('z_index', None)

    @property
    def class_name(self) -> Optional[str]:
        """A class name for the crosshair, particularly useful for styling the crosshair.
        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the crosshair. Defaults to :obj:`None <python:None>`, which
        applies ``'#cccccc' for datetime and numeric axes, and
        ``'rgba(204, 214, 235, 0.25)'`` for category axes (where the crosshair by default
        highlights the whole category).

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def dash_style(self) -> Optional[str]:
        """Name of the dash style to use for the crosshair. Defaults to ``Solid``.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._dash_style

    @dash_style.setter
    def dash_style(self, value):
        if not value:
            self._dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'dash_style expects a '
                                                  f'recognized value, but received: '
                                                  f'{value}')
            self._dash_style = value

    @property
    def snap(self) -> Optional[bool]:
        """If ``True``, the crosshair should snap to the point. If ``False``, the
        crosshair will follow the pointer. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._snap

    @snap.setter
    def snap(self, value):
        if value is None:
            self._snap = None
        else:
            self._snap = bool(value)

    @property
    def width(self) -> Optional[int | float | Decimal]:
        """The width of the crosshair, in pixels. Defaults to :obj:`None <python:None>`,
        which applies ``1`` for numeric and datetime axes, and a single category width
        for category axes.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = validators.numeric(value, allow_empty = True)

    @property
    def z_index(self) -> Optional[int | float | Decimal]:
        """The Z index for the crosshair. Defaults to ``2``.

        .. hint::

          Higher Z-indices allow drawing the crosshair on top of the series or behind the
          grid lines.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._z_index

    @z_index.setter
    def z_index(self, value):
        self._z_index = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'dash_style': as_dict.get('dashStyle', None),
            'snap': as_dict.get('snap', None),
            'width': as_dict.get('width', None),
            'z_index': as_dict.get('zIndex', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'className': self.class_name,
            'color': self.color,
            'dashStyle': self.dash_style,
            'snap': self.snap,
            'width': self.width,
            'zIndex': self.z_index
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.utility_classes.date_time_label_formats import DateTimeLabelFormats as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'day': 'test',
      'hour': 'test',
      'millisecond': 'test',
      'minute': 'test',
      'month': 'test',
      'second': 'test',
      'week': 'test',
      'year': 'test'
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/date_time_label_formats/01.js', False, None),

    ('utility_classes/date_time_label_formats/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('utility_classes/date_time_label_formats/01.js', True, None),

    ('utility_classes/date_time_label_formats/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.global_options.language.accessibility import AccessibilityLanguageOptions as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'announce_new_data': {
          'newDataAnnounce': 'New Data!',
          'newPointAnnounceMultiple': 'New data points added',
          'newPointAnnounceSingle': 'New data point added',
          'newSeriesAnnounceMultiple': 'New data series added',
          'newSeriesAnnounceSingle': 'Added a data series'
      },
      'axis': {
          'rangeCategories': 'Categories',
          'rangeFromTo': 'X to Y',
          'timeRangeDays': 'Monday - Friday',
          'timeRangeHours': '0h - 24h',
          'timeRangeMinutes': ':00 - :59',
          'timeRangeSeconds': ':00:00 - :00:59',
          'xAxisDescriptionPlural': 'X Axis Description when pluralized',
          'xAxisDescriptionSingular': 'X Axis Description',
          'yAxisDescriptionPlural': 'Y-Axis description when pluralized',
          'yAxisDescriptionSingular': 'Y-Axis Description'
      },
      'chart_container_label': 'some string',
      'chart_types': {
          'barMultiple': 'Bar Charts',
          'barSingle': 'Bar Chart',
          'boxplotMultiple': 'Boxplot Charts',
          'boxplotSingle': 'Boxplot Chart',
          'bubbleMultiple': 'Bubble Charts',
          'bubbleSingle': 'Bubble Chart',
          'columnMultiple': 'Column Charts',
          'columnSingle': 'Column Chart',
          'combinationChart': 'Combo Chart',
          'defaultMultiple': 'Charts',
          'defaultSingle': 'Chart',
          'emptyChart': 'Empty Chart',
          'lineMultiple': 'Line Charts',
          'lineSingle': 'Line Chart',
          'mapTypeDescription': 'Map',
          'pieMultiple': 'Pie Charts',
          'pieSingle': 'Pie Chart',
          'scatterMultiple': 'Scatter Charts',
          'scatterSingle': 'Scatter Chart',
          'splineMultiple': 'Spline Charts',
          'splineSingle': 'Spline Chart',
          'unknownMap': 'Unknown Map'
      },
      'credits': 'credits go here',
      'default_chart_title': 'default title goes here',
      'drillup_button': 'drillup button label',
      'exporting': {
          'chartMenuLabel': 'Menu',
          'menuButtonLabel': 'Menu Button'
      },
      'graphic_container_label': 'Graphic Container',
      'legend': {
          'legendItem': 'Item',
          'legendLabel': 'Legend',
          'legendLabelNoTitle': 'Label with no Title'
      },
      'range_selector': {
          'clickButtonAnnouncement': 'Selected Range',
          'dropdownLabel': 'Range',
          'maxInputLabel': 'Max.',
          'minInputLabel': 'Min.'
      },
      'screen_reader_section': {
          'afterRegionLabel': 'After Region',
          'annotations': {
              'descriptionMultiplePoints': 'Description of multiple points',
              'descriptionNoPoints': 'Description with no points',
              'descriptionSinglePoint': 'Description of a single point',
              'heading': 'My Heading'
          },
          'beforeRegionLabel': 'Before Region',
          'endOfChartMarker': 'End of Chart'
      },
      'sonification': {
          'playAsSoundButtonText': 'Play as Sound',
          'playAsSoundClickAnnouncement': 'Playing as Sound'
      },
      'svg_container_label': 'SVG',
      'svg_container_title': 'SVG',
      'table': {
          'tableSummary': 'Summary goes here',
          'viewAsDataTableButtonText': 'View as Table'
      },
      'thousands_separator': ',',
      'zoom': {
          'mapZoomIn': 'Zoom In',
          'mapZoomOut': 'Zoom out',
          'resetZoomButton': 'Reset Zoom'
      }
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('global_options/language/accessibility/accessibility/01.js', False, None),

    ('global_options/language/accessibility/accessibility/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('global_options/language/accessibility/accessibility/01.js', True, None),

    ('global_options/language/accessibility/accessibility/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional
from decimal import Decimal

from validator_collection import validators, checkers

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.series.data.base import DataBase
from highcharts_core.options.plot_options.drag_drop import DragDropOptions
from highcharts_core.utility_classes.data_labels import DataLabel


class ConnectionBase(DataBase):
    """Data point that represents a connection ``from`` one data point, ``to`` another."""

    def __init__(self, **kwargs):
        self._from_ = None
        self._to = None

        self.from_ = kwargs.get('from_', None)
        self.to = kwargs.get('to', None)

        super().__init__(**kwargs)

    @property
    def from_(self) -> Optional[str]:
        """The node that the link runs from. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._from_

    @from_.setter
    def from_(self, value):
        self._from_ = validators.string(value, allow_empty = True)

    @property
    def to(self) -> Optional[str]:
        """The node that the link runs to. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._to

    @to.setter
    def to(self, value):
        self._to = validators.string(value, allow_empty = True)

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, cls):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls()
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a Connection Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank',
                                      None) or as_dict.get('labelRank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'from_': as_dict.get('from', None),
            'to': as_dict.get('to', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'from': self.from_,
            'to': self.to,

            'accessibility': self.accessibility,
            'className': self.class_name,
            'color': self.color,
            'colorIndex': self.color_index,
            'custom': self.custom,
            'description': self.description,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
            'selected': self.selected,
        }

        return untrimmed


class ConnectionData(ConnectionBase):
    """Data point that represents a connection ``from`` one data point, ``to`` another."""

    def __init__(self, **kwargs):
        self._data_labels = None
        self._drag_drop = None

        self.data_labels = kwargs.get('data_labels', None)
        self.drag_drop = kwargs.get('drag_drop', None)

        super().__init__(**kwargs)

    @property
    def data_labels(self) -> Optional[DataLabel]:
        """Individual data label for the data point.

        :rtype: :class:`DataLabel` or :obj:`None <python:None>`
        """
        return self._data_labels

    @data_labels.setter
    @class_sensitive(DataLabel)
    def data_labels(self, value):
        self._data_labels = value

    @property
    def drag_drop(self) -> Optional[DragDropOptions]:
        """The draggable-points module allows points to be moved around or modified in the
        chart.

        In addition to the options mentioned under the dragDrop API structure, the module
        fires three (JavaScript) events:

          * ``point.dragStart``
          * ``point.drag``
          * ``point.drop``

        :rtype: :class:`DragDropOptions` or :obj:`None <python:None>`
        """
        return self._drag_drop

    @drag_drop.setter
    @class_sensitive(DragDropOptions)
    def drag_drop(self, value):
        self._drag_drop = value

    @classmethod
    def from_array(cls, value):
        """Generator method which produces a collection of :class:`ConnectionData`
        instances derived from ``value``. Generally consumed by the setter methods in
        series-type specific data classes.

        :rtype: :class:`list <python:list>` of :obj:`ConnectionData` instances
        """
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'ConnectionData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls()
            elif len(item) == 2:
                as_obj = cls(from_ = item[0],
                             to = item[1])
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a Connection Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank',
                                      None) or as_dict.get('labelRank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'from_': as_dict.get('from', None),
            'to': as_dict.get('to', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'dataLabels': self.data_labels,
            'dragDrop': self.drag_drop,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class WeightedConnectionData(ConnectionData):
    """Variant of :class:`ConnectionData` that also applies a ``weight`` to the
    connection."""

    def __init__(self, **kwargs):
        self._weight = None

        self.weight = kwargs.get('weight', None)

        super().__init__(**kwargs)

    @property
    def weight(self) -> Optional[int | float | Decimal]:
        """The weight of the link. Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._weight

    @weight.setter
    def weight(self, value):
        self._weight = validators.numeric(value, allow_empty = True)

    @classmethod
    def from_array(cls, value):
        """Generator method which produces a collection of :class:`ConnectionData`
        instances derived from ``value``. Generally consumed by the setter methods in
        series-type specific data classes.

        :rtype: :class:`list <python:list>` of :obj:`ConnectionData` instances
        """
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'ConnectionData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls()
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a Connection Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank',
                                      None) or as_dict.get('labelRank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'from_': as_dict.get('from', None),
            'to': as_dict.get('to', None),

            'weight': as_dict.get('weight', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'weight': self.weight,

            'dataLabels': self.data_labels,
            'dragDrop': self.drag_drop,

            'from': self.from_,
            'to': self.to,

            'accessibility': self.accessibility,
            'className': self.class_name,
            'color': self.color,
            'colorIndex': self.color_index,
            'custom': self.custom,
            'description': self.description,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
            'selected': self.selected
        }

        return untrimmed


class OutgoingWeightedConnectionData(WeightedConnectionData):
    """Variant of :class:`WeightedConnectionData` that supports the ``outoging`` flag."""

    def __init__(self, **kwargs):
        self._outgoing = None

        self.outgoing = kwargs.get('outgoing', None)

        super().__init__(**kwargs)

    @property
    def outgoing(self) -> Optional[bool]:
        """If ``True``, indicates that the data point links out of the system. Defaults
        to :obj:`None <python:None>`, which behaves as ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._outgoing

    @outgoing.setter
    def outgoing(self, value):
        if value is None:
            self._outgoing = None
        else:
            self._outgoing = bool(value)

    @classmethod
    def from_array(cls, value):
        """Generator method which produces a collection of
        :class:`OutgoingWeightedConnectionData` instances derived from ``value``.
        Generally consumed by the setter methods in series-type specific data classes.

        :rtype: :class:`list <python:list>` of :obj:`OutgoingWeightedConnectionData`
          instances
        """
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'OutgoingWeightedConnectionData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls()
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a Connection Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'outgoing': as_dict.get('outgoing', None),

            'weight': as_dict.get('weight', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'from_': as_dict.get('from', None),
            'to': as_dict.get('to', None),

            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank',
                                      None) or as_dict.get('labelRank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'outgoing': self.outgoing
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



"""Implements decorators used throughout the library."""
import json
from functools import wraps
from collections import UserDict

from validator_collection import checkers

from highcharts_core import errors, constants


def validate_types(value,
                   types = None,
                   allow_dict = True,
                   allow_json = True,
                   allow_none = True,
                   allow_js_literal = True,
                   force_iterable = False,
                   function_name = None):
    """Validates that ``value`` is one or more of the allowed types, where the first
    type passed in ``types`` is the primary type that it will be returned as.

    :param value: The value to be validated.
    :type value: Any

    :param types: :class:`type <python:type>` object or iterable of
      :class:`type <python:type>` objects used to indicate which types are allowed. The
      first (or only) item in ``types`` will indicate the primary type that ``value`` will
      be returned as.
    :types: :class:`type <python:type>` or iterable of :class:`type <python:type>`

    :param allow_dict: If ``True``, will accept a :class:`dict <python:dict>` object as
      ``value``. Defaults to ``True``.
    :type allow_dict: :class:`bool <python:bool>`

    :param allow_json: If ``True``, will accept a :class:`str <python:str>` object as
      ``value``, under the assumption that is is deserializable as JSON. Defaults to
      ``True``.
    :type allow_json: :class:`bool <python:bool>`

    :param allow_none: If ``True``, will accept an empty or :obj:`None <python:None>`
      object as ``value``. Defaults to ``True``.
    :type allow_none: :class:`bool <python:bool>`

    :param allow_js_literal: If ``True``, will accept a :class:`str <python:str>` object
      as ``value``, under the assumption that it is deserializable as a JavaScript object
      literal. Defaults to ``True``.
    :type allow_js_literal: :class:`bool <python:bool>`

    :param force_iterable: If ``True``, will accept an iterable object as ``value``.
      Defaults to ``False`` (because most attributes are just singletons).
    :type force_iterable: :class:`bool <python:bool>`

    :param function_name: The optional name of the function that was originally called.
    :type function_name: :class:`str <python:str>`

    :returns: ``value`` de-serialized to the primary type (first
      :class:`type <python:type>` in ``types``)
    :rtype: first :class:`type <python:type>` in ``types``

    :raises HighchartsImplementationError: if ``types`` is empty

    :raises HighchartsValueError: if ``types`` does not contain a
      :class:`type <python:type>` or iterable of :class:`type <python:type>` objects

    :raises HighchartsValueError: if the primary type does not conform to the
      :class:`HighchartsMeta` interface definition

    """
    if not types:
        raise errors.HighchartsImplementationError('types cannot be empty - must be a type or '
                                         'iterable of types')

    if not types:
        raise errors.HighchartsImplementationError('types cannot be empty - must be a type or '
                                         'iterable of types')

    try:
        types_list = [x for x in types]
    except TypeError:
        types_list = [types]

    for item in types_list:
        if not isinstance(item, type):
            raise errors.HighchartsValueError(f'types must contain one or more type '
                                              f'objects. Received a {type(item)}.')

    primary_type = types_list[0]
    if not hasattr(primary_type, 'from_js_literal'):
        allow_js_literal = False

    if allow_none and force_iterable and checkers.is_iterable(value) and not value:
        value = []
    elif allow_none and isinstance(value, constants.EnforcedNullType):
        pass
    elif allow_none and not value:
        value = None
    elif not allow_none and not value:
        raise errors.HighchartsValueError('value is not expected to be empty, but was '
                                          'empty')
    elif allow_dict and isinstance(value, dict):
        try:
            value = primary_type.from_dict(value)
        except AttributeError as error:
            if not hasattr(primary_type, 'from_dict'):
                raise errors.HighchartsValueError(f'supplied type '
                                                  f'({primary_type.__name__}) '
                                                  f'does not conform to the '
                                                  f'HighchartsMeta interface')
            else:
                raise error
    elif allow_js_literal and isinstance(value, str):
        try:
            value = primary_type.from_js_literal(value)
        except ValueError:
            pass

    if (
        force_iterable and
        checkers.is_iterable(value, forbid_literals = (str, dict, bytes, UserDict)) and
        hasattr(primary_type, 'from_array')
    ):
        value = primary_type.from_array(value)
    elif allow_json and isinstance(value, (str, bytes)):
        try:
            value = primary_type.from_json(value)
        except AttributeError:
            raise errors.HighchartsValueError(f'supplied type '
                                              f'({primary_type.__class__.__name__}) '
                                              f'does not conform to the '
                                              f'HighchartsMeta interface')
        except TypeError:
            value = json.loads(value)

    if force_iterable and checkers.is_iterable(value):
        value = [validate_types(x,
                                types = types,
                                allow_dict = allow_dict,
                                allow_json = allow_json,
                                allow_js_literal = allow_js_literal,
                                force_iterable = force_iterable)
                 for x in value]
    elif allow_none and isinstance(value, constants.EnforcedNullType):
        pass
    elif allow_none and not value:
        pass
    elif not isinstance(value, primary_type):
        error_string = f'expects a {primary_type.__name__}'
        if function_name:
            error_string = f'{function_name} ' + error_string

        allow_js = allow_json or allow_js_literal
        if allow_dict and allow_js and force_iterable and allow_none:
            error_string += ', dict, str, iterable, or empty object.'
        elif allow_dict and allow_js and force_iterable:
            error_string += ', dict, str, or iterable object.'
        elif allow_dict and allow_js and allow_none:
            error_string += ', dict, str, or empty object.'
        elif allow_dict and force_iterable and allow_none:
            error_string += ', dict, iterable, or empty object.'
        elif allow_dict and allow_js:
            error_string += ', dict, or str object.'
        elif allow_dict and force_iterable:
            error_string += ', dict, or iterable object.'
        elif allow_dict and allow_none:
            error_string += ', dict, or empty object.'
        elif allow_js and force_iterable:
            error_string += ', str, or iterable object.'
        elif allow_js and allow_none:
            error_string += ', str, or empty object.'
        elif force_iterable and allow_none:
            error_string += ', iterable, or empty object.'
        elif allow_js:
            error_string += ' or str object.'
        elif force_iterable:
            error_string += ' or iterable object.'
        elif allow_none:
            error_string += ' or empty object.'

        error_string += f' Received {value.__class__.__name__}.'

        raise errors.HighchartsValueError(error_string)

    return value


def class_sensitive(types = None,
                    allow_dict = True,
                    allow_json = True,
                    allow_none = True,
                    allow_js_literal = True,
                    force_iterable = False):
    """Validates that the values passed to a decorated function or method are
    de-serialized to the appropriate type.

    :param types: :class:`type <python:type>` object or iterable of
      :class:`type <python:type>` objects used to indicate which types are allowed. The
      first (or only) item in ``types`` will indicate the primary type that ``value`` will
      be returned as.
    :types: :class:`type <python:type>` or iterable of :class:`type <python:type>`

    :param allow_dict: If ``True``, will accept a :class:`dict <python:dict>` object as
      ``value``. Defaults to ``True``.
    :type allow_dict: :class:`bool <python:bool>`

    :param allow_json: If ``True``, will accept a :class:`str <python:str>` object as
      ``value``, under the assumption that is is deserializable as JSON. Defaults to
      ``True``.
    :type allow_json: :class:`bool <python:bool>`

    :param allow_none: If ``True``, will accept an empty or :obj:`None <python:None>`
      object as ``value``. Defaults to ``True``.
    :type allow_none: :class:`bool <python:bool>`

    :param allow_js_literal: If ``True``, will accept a :class:`str <python:str>` object
      as ``value``, under the assumption that it is deserializable as a JavaScript object
      literal. Defaults to ``True``.
    :type allow_js_literal: :class:`bool <python:bool>`

    :param force_iterable: If ``True``, will accept an iterable object as ``value``.
      Defaults to ``False`` (because most attributes are just singletons).
    :type force_iterable: :class:`bool <python:bool>`

    .. note::

      To apply the decorator to a property setter method (the most-common use case), place
      it *after* the ``@<property name>.setter`` decorator and directly above the function
      name like so:

      .. code-block:: python

        @some_property.setter
        @class_sensitive(...)
        def some_property(self, value):
            ...

    :returns: The result of the decorated function or method having validated the class
      typing.

    :raises HighchartsImplementationError: if ``types`` is empty

    :raises HighchartsValueError: if ``types`` does not contain a
      :class:`type <python:type>` or iterable of :class:`type <python:type>` objects

    :raises HighchartsValueError: if the primary type does not conform to the
      :class:`HighchartsMeta` interface definition

    """
    def decorator(func):
        @wraps(func)
        def func_wrapper(*args,
                         **kwargs):
            function_name = func.__name__

            try:
                value = args[1]
            except IndexError:
                raise errors.HighchartsError('Something went wrong. Unsure how this '
                                             'might happen.')

            value = validate_types(value,
                                   types = types,
                                   allow_dict = allow_dict,
                                   allow_json = allow_json,
                                   allow_none = allow_none,
                                   allow_js_literal = allow_js_literal,
                                   force_iterable = force_iterable,
                                   function_name = function_name)

            result = func(args[0], value)

            return result

        return func_wrapper

    return decorator
